﻿/*
*************************************************************************************************  
*                  Title: autofac 各种用法演示
*            Description: autofac的用法太多了，但其中最常用的只有1种，示例RegisterModuleByXML1   
*              Copyright: Copyright@2014 XXXX Technologies Co., Limited
*                Company: XXXX Technologies
*                 Author: Taylor.Tian
*            Create Date: 21/11/2014
*          Last Modifier:    
*       Last Modify Date:    
*                Version: 1.00.00
*************************************************************************************************
*/
using System;
using System.Collections.Generic;

using Autofac;
using Autofac.Configuration;
using Autofac.Core;

namespace AutofacDemo
{
    public class AutofacUsage
    {
        public ContainerBuilder Builder { get; set; }

        public AutofacUsage()
        {
            //Builder = new ContainerBuilder();
        }

        /// <summary>
        /// 注册一个实例化的对象
        /// </summary>
        public void RegisterInstance1()
        {
            //重新实例化ContainerBuilder
            Builder = new ContainerBuilder();

            //Autofac框架控制反转（实例化类型）
            Builder.RegisterInstance(new Cat(10, "Tom")).As<IAnimal>();

            //获取IOC容器
            var container = Builder.Build();

            //由容器注入
            var instance = container.Resolve<IAnimal>();

            //测试实例中的方法
            PrintInstance(instance);
        }

        /// <summary>
        /// 这也是注册一个实例化的对象
        /// </summary>
        public void RegisterInstance2()
        {
            Builder = new ContainerBuilder();
            Builder.Register(c => new Cat(11, "Tom")).As<IAnimal>();

            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }

        /// <summary>
        /// 这也是注册一个实例化的对象,更有意思的是它可以根据条件，注册不同的实例化对象
        /// </summary>
        public void RegisterInstance3()
        {
            //Builder = new ContainerBuilder();
            //Builder.Register(c => new Mouse(11, "Tom")).As<Mouse>();
            //var container = Builder.Build();
            int condition = 10;
            Builder = new ContainerBuilder();
            Builder.Register<Cat>(
                (c, p) =>
                {
                    //foreach (var item in p)
                    //    System.Diagnostics.Trace.WriteLine(item);
                    //var age = p.Named<int>("age");
                    //var name = p.Named<string>("name");
                    if (condition > 3)
                        return new Cat(1, "Tom");
                    else
                        return new SmallCat(2, "Jack");
                }
                ).As<IAnimal>(); ;
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }

        /// <summary>
        /// 注册一个无参构造函数的类
        /// </summary>
        public void RegisterTypeNoneParameterConstructor()
        {
            Builder = new ContainerBuilder();
            Builder.RegisterType<Cat>().As<IAnimal>();
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }

        /// <summary>
        /// 注册一个有参构造函数(1个参数):使用参数名字
        /// </summary>
        public void RegisterTypeIncludeOneParameterConstructor1()
        {
            Builder = new ContainerBuilder();
            //使用参数名字
            Builder.RegisterType<Cat>().WithParameter("age", 1).As<IAnimal>();
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }

        /// <summary>
        /// 注册一个有参构造函数(1个参数):使用参数类型
        /// </summary>
        public void RegisterTypeIncludeOneParameterConstructor2()
        {
            Builder = new ContainerBuilder();
            //使用参数类型
            Builder.RegisterType<Cat>().WithParameter(new TypedParameter(typeof(int), 2)).As<IAnimal>();
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }


        /// <summary>
        /// 注册一个有参构造函数(1个参数):使用Lambda
        /// </summary>
        public void RegisterTypeIncludeOneParameterConstructor3()
        {
            Builder = new ContainerBuilder();
            //使用Lambda
            Builder.RegisterType<Cat>().WithParameter(new ResolvedParameter(
                (p, c) => p.ParameterType == typeof(int) && p.Name == "age",
                (p, c) => 3)
                ).As<IAnimal>();
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }

        /// <summary>
        /// 注册一个有参构造函数(2个参数)
        /// </summary>
        public void RegisterTypeIncludeTwoParametersConstructor()
        {
            Builder = new ContainerBuilder();
            List<ResolvedParameter> list = new List<ResolvedParameter>();
            list.Add(
                new ResolvedParameter(
                    (p, c) => p.ParameterType == typeof(int) && p.Name == "age",
                    (p, c) => 4
                    )
                );
            list.Add(
                new ResolvedParameter(
                    (p, c) => p.ParameterType == typeof(int) && p.Name == "name",
                    (p, c) => "Jim"
                    )
                );
            Builder.RegisterType<Cat>().WithParameters(list).As<IAnimal>();
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>();
            PrintInstance(instance);
        }

        /// <summary>
        /// 不以实例注册，仅指定选择构造函数(其实这种更常用)
        /// </summary>
        public void RegisterTypeCommon1()
        {
            Builder = new ContainerBuilder();
            //使用参数类型,还可以加一个属性
            Builder.RegisterType<Cat>().UsingConstructor(typeof(int), typeof(string)).WithProperty("Address", "xi'an").As<IAnimal>(); ;
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>(new NamedParameter("age", 12), new NamedParameter("name", "Tom"));
            PrintInstance(instance);
        }

        /// <summary>
        /// 这个也可以
        /// </summary>
        public void RegisterTypeCommon2()
        {
            Builder = new ContainerBuilder();
            Builder.Register((c, p) => new Cat(p.Named<int>("age"), p.Named<string>("name"))).As<IAnimal>();
            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>(new NamedParameter("age", 12), new NamedParameter("name", "Tom"));
            PrintInstance(instance);
        }

        /// <summary>
        /// 注册泛型
        /// </summary>
        public void RegisterTypeCommon3()
        {
            Builder = new ContainerBuilder();
            Builder.RegisterGeneric(typeof(CGenericType<,>)).As(typeof(ICGenericType<,>));
            var container = Builder.Build();

            Type genericType = typeof(ICGenericType<,>);
            Type instance = genericType.MakeGenericType(
                new Type[] { typeof(int), typeof(string) }
                );

            var obj = container.Resolve(instance) as ICGenericType<int, string>;
            var output = obj.GetInfor();
        }

        /// <summary>
        /// 这个更厉害，注册程序集
        /// </summary>
        public void RegisterAssembly()
        {
            var assembly = typeof(Cat).Assembly;
            Builder = new ContainerBuilder();
            Builder.RegisterAssemblyTypes(assembly);
            //Builder.RegisterAssemblyTypes(assembly).Where(type => type.Name == "Cat");

            var container = Builder.Build();
            bool isRegistered = container.IsRegistered<Cat>();
            var instance = container.Resolve<Cat>(new NamedParameter("age", 11));
            PrintInstance(instance);
        }

        /// <summary>
        /// 注册module-code方式,如果要读取xml，注意在app.config中配置runtime节点
        /// </summary>
        public void RegisterModule()
        {
            Builder = new ContainerBuilder();
            Builder.RegisterModule(new MyModule());

            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>(new NamedParameter("age", 10));
            PrintInstance(instance);
        }

        /// <summary>
        /// 最常用的：注册module-xml方式
        /// 上述方式都不是最常用的，最常用的方式是这个xml configuration
        /// </summary>
        public void RegisterModuleByXML1()
        {
            string autofacConfigFile = "Autofac.config1.xml";
            Builder = new ContainerBuilder();

            var xmlReader = new XmlFileReader(autofacConfigFile);
            //var reader = new ConfigurationSettingsReader("myautofac");//app.config
            Builder.RegisterModule(xmlReader);
            var container = Builder.Build();

            var instance = container.Resolve<IAnimal>(new NamedParameter("age", 10));
            PrintInstance(instance);

            Type genericType = typeof(ICGenericType<,>);
            Type genericInstance = genericType.MakeGenericType(
                new Type[] { typeof(int), typeof(string) }
                );

            var obj = container.Resolve(genericInstance) as ICGenericType<int, string>;
            var output = obj.GetInfor();
        }

        /// <summary>
        /// 注册module-xml方式,包含了RegisterModuleByXML1
        /// 这个混合了xml configuration和code注册2种在一起
        /// </summary>
        public void RegisterModuleByXML2()
        {
            string autofacConfigFile = "Autofac.config2.xml";
            Builder = new ContainerBuilder();

            var xmlReader = new XmlFileReader(autofacConfigFile);
            //var reader = new ConfigurationSettingsReader("myautofac");//app.config
            Builder.RegisterModule(xmlReader);

            var container = Builder.Build();
            var instance = container.Resolve<IAnimal>(new NamedParameter("age", 10));
            PrintInstance(instance);

            Type genericType = typeof(ICGenericType<,>);
            Type genericInstance = genericType.MakeGenericType(
                new Type[] { typeof(int), typeof(string) }
                );

            var obj = container.Resolve(genericInstance) as ICGenericType<int, string>;
            var output = obj.GetInfor();
        }

        public void PrintInstance(IAnimal instance)
        {
            Tuple<int, string, string> output = instance.Cry();
            System.Diagnostics.Trace.WriteLine(string.Format("{0}-{1}-{2}", output.Item1, output.Item2, output.Item3));
        }
    }

    #region 注册目标
    public interface IAnimal
    {
        Tuple<int, string, string> Cry();
    }

    public class Cat : IAnimal
    {
        public int Age { get; set; }
        public string Name { get; set; }
        public string Address { get; set; }

        public Cat()
        {
            Age = 0;
            Name = string.Empty;
        }

        public Cat(int age)
        {
            Age = age;
            Name = string.Empty;
        }

        public Cat(int age, string name)
        {
            Age = age;
            Name = name;
        }


        public Tuple<int, string, string> Cry()
        {
            return Tuple.Create<int, string, string>(Age, Name, "Cat is here!" + Address);
        }
    }

    public class SmallCat : Cat
    {
        public SmallCat(int age, string name)
            : base(age, name)
        {

        }
    }

    public class Mouse : IAnimal
    {
        public int Age { private get; set; }
        public string Name { private get; set; }

        public Mouse(int age, string name)
        {
            Age = age;
            Name = name;
        }

        public Tuple<int, string, string> Cry()
        {
            return Tuple.Create<int, string, string>(Age, Name, "Mouse is here!");
        }
    }

    public interface ICGenericType<T1, T2>
    {
        string GetInfor();
    }

    public class CGenericType<T1, T2> : ICGenericType<T1, T2>
    {
        T1 Age { get; set; }
        T2 Name { get; set; }

        public CGenericType()
        {
            Age = default(T1);
            Name = default(T2);
        }

        public string GetInfor()
        {
            return string.Format("The infor as below {0}:{1}", Name, Age);
        }
    }

    public class MyModule : Module
    {
        protected override void Load(ContainerBuilder builder)
        {
            builder.Register(c => new Cat(11) { Name = "Tom" });
            builder.Register(c => new Mouse(22, "Jim"));
        }
    }

    public class XMLModule : Module
    {
        public int Num { get; set; }

        public XMLModule(int count)
        {
            Num = count;
        }

        protected override void Load(ContainerBuilder builder)
        {
            builder.Register(c => new Cat(11) { Name = "Tom" }).As<IAnimal>();
            builder.Register(c => new Mouse(22, "Jim")).As<IAnimal>();

            //generic type
            builder.RegisterGeneric(typeof(CGenericType<,>)).As(typeof(ICGenericType<,>));
        }
    }
    #endregion
}
